import types
from copy import deepcopy
from django.db import models

class ExpanderField(object):
    """
    This type of field can be used to include some common fields and methods in models.
    It take a class with those common things as parameter in the constructor and will 
    add all fields in that class to models classes in which a field of
    this class type is declared. If the class containing common fields 
    have fields of type: ForeignKey and ManyToManyField, then the related_name attribute
    of those fields will be renamed to a value of the form: model_name_original_value .
    Ej: if related_name = child_set and model name is MenuItem, it will be renamed to 
    MenuItem_child_set. To add methods of the common class you must inherit from it.
    An importan thing to note here is that you must inherit first from the common class, 
    order matters.

    Use case:
    class ContentManager(models.Manager):
        pass
        
    class CommonFields:
        pub_date = models.DateTimeField()
        created_by = models.ForeignKey(User, related_name = 'created_by_set')
        last_modified_by = models.ForeignKey(User, related_name = 'last_modified_by_set')

        objects = ContentManager()

        def save(self):
            #do something
            pass
            
    class NewsItem(CommonFields, models.Model):
        title = models.CharField(maxlength = 100)
        body = models.CharField(maxlength = 200)
        common = ExpanderField(CommonFields)

    this will create a class of this form:

    class NewsItem(models.Model):
        title = models.CharField(maxlength = 100)
        body = models.CharField(maxlength = 200)
        pub_date = models.DateTimeField()
        created_by = models.ForeignKey(User, related_name = 'created_by_set')
        last_modified_by = models.ForeignKey(User, related_name = 'last_modified_by_set')

        objects = ContentManager()

    """
    def __init__(self, field_container_class):
        self.field_container_class = field_container_class

    def contribute_to_class(self, cls, name):
        attr_list = [attr for attr in dir(self.field_container_class) 
                     if attr not in  ('__doc__', '__module__', '__class__', '__dict__')]
        container = self.field_container_class
        
        for attr in attr_list:
            clone = None
            attr_value = getattr(container, attr)
            
            if type(attr_value) != types.MethodType:
                clone = deepcopy(attr_value)
                
                if (isinstance(clone, models.ForeignKey) or 
                    isinstance(clone, models.ManyToManyField)) and \
                    clone.rel.related_name is not None:
                    clone.rel.related_name = cls.__name__ + '_' + clone.rel.related_name
                
            if clone is not None:
                cls.add_to_class(attr, clone)



"""
This class handles conversions to and from CMDBuild's own JSON format.

During encoding, lists are converted to "collections", and basic types
to a dictionary explicitly declaring the base type itself.

This may be extended in the future to talk more precisely to CMDBuild,
once the need arises and the format is decoded.
"""

from datetime import date

class Bean(object):
	"""
	Encapsulates objects to be recognised by CMDBuild's deserializer.
	"""
	def __init__(self, **kwargs):
		self._map = dict(**kwargs)
	def json_map(self):
		"""
		Advertises the bean's original class and returns the property dictionary
		"""
		serializable_map = self._map.copy()
		if hasattr(self, 'original_class'):
			serializable_map['json.converter.original.class'] = self.original_class
		return serializable_map

class AssetBean(Bean):
	original_class = 'cmdbuild.collection.AssetBean'

class ServerException(Exception):
	def __init__(self, e):
		if not 'JSONConverter.exception' in e:
			raise TypeError, '%s must be initialized with proper ' \
				'JSON-decoded response' % self.__class__.__name__
		self.exc = e
		super(ServerException, self).__init__(self._get_msg())
	def _get_msg(self):
		msg = ['Server-side exception: "%s" in class %s' % (
			self.exc['exception.message'],
			self.exc['exception.original.class'])]
		msg.extend(['%s.%s() in %s:%s' % (l['class.name'], 
			l['method.name'], l['file.name'],
			l['line.number']) for l in self.exc['stacktrace']])
		return "\n".join(msg)

def decorate(obj):
	"""
	Turns an object into a JSONConverter-palatable form *before* serialization.
	"""
	def add_type(obj, converter_type):
		"Converts a value into a map"
		return { 'json.converter.original.class': converter_type,
			'JSONConverter.value': obj }
	
	if hasattr(obj, 'json_map'):
		return obj.json_map()
	elif isinstance(obj, int):
		return add_type(obj, 'java.lang.Integer')
	elif isinstance(obj, basestring):
		return add_type(obj, 'java.lang.String')
	elif isinstance(obj, bool):
		return add_type(obj, 'java.lang.Boolean')
	elif isinstance(obj, date):
		# TODO: test
		from time import mktime
		return add_type(long(mktime(obj.timetuple()) * 1000), 'java.util.Date')
	elif isinstance(obj, (list, tuple)):
		conv_objs = [decorate(o) for o in obj]
		return {'JSONConverter.iscollection':
			True, 'JSONConverter.collection': conv_objs }
	elif isinstance(obj, dict):
		d = dict([(k, decorate(v)) for k,v in obj.items()])
		d['JSONConverter.map'] = True
		return d
	raise TypeError, "Don't know how serialize type %s (content: %s)" \
		% (type(obj), obj)

def clean_attributes(obj):
	"""
	Inside a dictionary, find all items with a NULL value and delete them.
	"""
	for name, value in obj.items():
		if value == 'json.converter.null':
			del obj[name]
	return obj

def strip(obj):
	"""
	Take CMDBuild's conversion output and strip it of its Java boring details;
	duck typing will do, for us.
	"""
	# Case 1: collection
	if 'JSONConverter.iscollection' in obj:
		return [strip(item) for item in obj['JSONConverter.collection']]
	if 'JSONConverter.map' in obj:
		del obj['JSONConverter.map']
		return dict([(k, strip(v)) for k,v in obj.items()])
	# Case 2: encoded class
	try:
		original_class = obj['json.converter.original.class']
		del obj['json.converter.original.class']
	except KeyError:
		original_class = None
	if original_class:
		# Act according to the class type
		if original_class in ('java.lang.String', 'java.lang.Boolean'):
			return obj['JSONConverter.value']
		elif original_class == 'java.util.Date':
			# Java returns this in milliseconds since the epoch
			return date.fromtimestamp(obj['JSONConverter.value']/1000)
		elif original_class in ('cmdbuild.collection.LookupBean',):
			return clean_attributes(obj)
		raise TypeError, "Cannot decode original class %s "\
			"(content: '%s')" % (original_class, obj)
	# Server exceptions
	if 'JSONConverter.exception' in obj:
		return ServerException(obj)
	# Last case: anything else
	raise TypeError, 'Cannot decode %s' % obj

def as_string(value):
	"""
	Converts a Python value into a string suitable for SQL-like representation.
	"""
	if value is False:
		return "false"
	elif value is True:
		return "true"
	elif value is None:
		return ""
	elif isinstance(value, int):
		return "%s" % value
	elif isinstance(value, basestring):
		return value
	elif isinstance(value, date):
		return value.isoformat()
	else:
		raise TypeError, "Can't convert data of type %s" % type(value)



from django.db import models
from django.utils.translation import ugettext_lazy as _

from django_cmdbuild.models.expander import ExpanderField
from django_cmdbuild.models.querysets import *
from django_cmdbuild.models import Lookup

def from_kwargs(func):
    def wrapper(*args, **kwargs):
        name, value = kwargs.items()[0]
        del kwargs[name]
        kwargs['name'] = name
        kwargs['value'] = value
        return func(*args, **kwargs)
    return wrapper

class MapFields:
    pass

class CMDBModelOptions(object):
    def save(self, *args, **kwargs):
        """
        Saves a CMDbuild instance, taking care to set the ``status`` flag
        and to get the right default value by polling the DB.
        """
        if hasattr(self, 'status'):
            self.status = 'A'
        if self._meta.has_auto_field:
            from django.utils.encoding import smart_unicode
            pk_val = self._get_pk_val()
            pk_set = pk_val is not None and smart_unicode(pk_val) != u''
            if not pk_set:
                from django.db import connection
                from introspection_common import query_class_catalog
                cursor = connection.cursor()
                sql = 'SELECT ' + query_class_catalog(self._meta.db_table, self._meta.pk.db_column)['default_value']
                cursor.execute(sql)
                r = cursor.fetchall()[0]
                setattr(self, self._meta.pk.attname, r[0])
        models.Model.save(self, *args, **kwargs)

    def __unicode__(self):
        if self._description:
            return self._description
        else:
            return unicode(self.pk)
   

    def _get_status_and_date(self):
        if self.status == 'A':
            r = _(u'Active (%s)')  % self.begin_date
            return r
        elif self.status == 'U':
            return _(u'Inactive (%s-%s)') % (
                self.begin_date, self.end_date)
        else:
            return self.status
    status_and_date = property(_get_status_and_date)

    def _get_end_date(self):
        """
        For an inactive instance, returns the timestamp of its closing.
        """
        # Check that we are not active
        if self.status == 'A':
            return None

        # Create the class on the fly; change the table; set the attribute
        class history_model(CMDBModelOptions, models.Model, ClassFields):
            _enddate = models.DateTimeField(db_column='EndDate', editable=False)
            commonfields = ExpanderField(ClassFields)
            class Meta:
                db_table = self._meta.db_table + '_history'

        # Set the manager
        #history_model.objects.contribute_to_class(history_model, objects)

        # Retrieve the instance
        instance = history_model.objects.get(pk=self.pk)

        return instance._enddate
    end_date = property(fget=_get_end_date)

    @classmethod
    def _make_bean(cls, name, value):
        """
        Take an attribute name and value (for this class), and return an
        AssetBean that can be serialized and provided to CMDBuild over
        a remote API call.
        """
        from django_cmdbuild.serializer.jsonconverter import as_string, AssetBean
        # If it is an object, use its ID
        if isinstance(value, models.Model):
            value = value.pk
        return AssetBean(name=cls._meta.get_field(name).db_column,
            value=as_string(value))

    @classmethod
    def _get_cmdbuild_attributemap(cls, attributes):
        """
        Take a dictionary of attribute/value pairs and turn the Django instance
        attribute names into CMDBuild attribute names.
        """
        # Turn objects into ids and turn names into column names
        _map = {}
        for name, value in attributes.items():
            if isinstance(value, models.Model):
                value = value.pk
            _map[cls._meta.get_field(name).db_column] = value
        return _map

    def _get_full_attributemap(self):
        """
        Returns all attributes in a map. Mostly useful for debugging.
        """
        return dict([(f.attname, getattr(self, f.attname))
            for f in self._meta.fields])

    def _reload(self):
        """
        Gets the instance again from the DB and copies the field
        attributes.  Ugly, but works.
        """
        updated = self.__class__.objects.get(pk=self.pk)
        for field in self._meta.fields:
            key = field.attname
            setattr(self, key, getattr(updated, key))

class CodeField(models.CharField):
    def __init__(self, *args, **kwargs):
        kwargs.setdefault('max_length', 100)
        kwargs.setdefault('db_column', 'Code')
        kwargs.setdefault('blank', True)
        kwargs.setdefault('null', True)
        super(CodeField, self).__init__(*args, **kwargs)
    def get_internal_type(self):
        return "CharField"

class DescriptionField(models.CharField):
    def __init__(self, *args, **kwargs):
        kwargs.setdefault('max_length', 250)
        kwargs.setdefault('db_column', 'Description')
        super(DescriptionField, self).__init__(*args, **kwargs)
    def get_internal_type(self):
        return "CharField"
    def contribute_to_class(self, cls, name):
        def read_description(self):
            return getattr(self, name)
        setattr(cls, '_description', property(read_description))
        super(DescriptionField, self).contribute_to_class(cls, name)

class LookupDescriptor(object):
	def __init__(self, attribute_name):
		self.attribute_name = attribute_name
	

class LookupField(models.IntegerField):
	def get_internal_type(self):
		return "IntegerField"
	def __init__(self, lookup_type, *args, **kwargs):
		self.lookup_type = lookup_type
		kwargs.setdefault('choices', Lookup.objects.choices(lookup_type))
		super(LookupField, self).__init__(*args, **kwargs)
	def get_lookup_obj(self, number):
		return Lookup.objects.get_by_number(self.lookup_type, number)
	def get_lookup_label(self, number):
		return self.get_lookup_obj(number).description
	def contribute_to_class(self, cls, name):
		self.attribute_name = name
		super(LookupField, self).contribute_to_class(cls, name)
		setattr(cls, name + '_text', property(lambda instance:
			self.get_lookup_label(getattr(instance, name))))
		
class IdClassField(models.TextField):
    def __init__(self):
        pass
    opts = {'db_column': 'IdClass', 'editable': False}
    def contribute_to_class(self, cls, name):
        opts = self.opts.copy()
        opts['default'] = '"%s"' % cls._meta.db_table
        super(IdClassField, self).__init__(**opts)
        super(IdClassField, self).contribute_to_class(cls, name)
    def get_internal_type(self):
        return "TextField"

class ClassFields:
    id = models.AutoField(primary_key=True, db_column='Id')
    idclass = IdClassField()
    #code = models.CharField(max_length=100, db_column='Code', blank=True, null=True)
    #description = models.CharField(max_length=250, db_column='Description')
    status = models.CharField(max_length=1, db_column='Status', editable=False,
        choices=(('A', _('Active')), ('U', _('Inactive'))))
    user = models.CharField(max_length=20, db_column='User', blank=True, null=True)
    begin_date = models.DateTimeField(db_column='BeginDate', auto_now=True)
    objects = QSManager(ClassFieldsQuerySet)()
        



from django.db import models

def QSManager(the_queryset_class):
    class QSManagerType(models.Manager):
        queryset_class = the_queryset_class
        def __init__(self, active_only=True):
            super(QSManagerType, self).__init__()
            self.active_only = active_only
        def get_query_set(self):
            qs = self.__class__.queryset_class(self.model)
            if self.active_only:
                qs = qs.active()
            return qs
        def all(self):
            return self.get_query_set()
        def with_inactive(self):
            # Warning: this does not chain! TODO: document this clearly
            return self.__class__.queryset_class(self.model)
        def __getattr__(self, attr, *args):
            try:
                return super(QSManagerType, self).__getattr__(attr, *args)
            except AttributeError:
                if 'model' in self.__dict__:
                    return getattr(self.__class__.get_query_set(self), attr, *args)
                else:
                    raise 
    QSManagerType.__name__ = the_queryset_class.__name__ + 'Manager'
    return QSManagerType

class ClassFieldsQuerySet(models.query.QuerySet):
    def active(self):
        return self.filter(status__exact='A')




class QueryClassCatalog(object):
    """
    Populates a dict representing the structure of the CMDBuild
    database as it appears from the internal catalog itself.
    """
    def __init__(self):
        from django.db import connection

        # Initalize the dictionary
        self.classes = {}

        # Retrieve the catalog view's rows from the DB
        cursor = connection.cursor()
        cursor.execute('SELECT classname, attributename, classcomment, attributemode, attributedescription, attributelookup, '
               'attributenull, attributedefault '
               'FROM cmdbclasscatalog')
        print cursor.description

        def fetchall(cursor):
            bare_rows = cursor.fetchall()
            for row in bare_rows:
                yield dict(map(lambda seq: (cursor.description[seq[0]][0], seq[1]), 
                    enumerate(row)))

        # We have one row for each field (that is, 0..N rows for each table)
        for r in fetchall(cursor):
            # Make a new dict entry, if we don't have it
            if not r['classname'] in self.classes:
                self.classes[r['classname']] = {}

            self.classes[r['classname']][r['attributename']] = {
                'mode': r['classcomment'],
                'desc': r['attributedescription'],
                'lookup': r['attributelookup'],
                'null': r['attributenull'] == '',
                'default_value': r['attributedefault']
             }

        cursor.execute('SELECT DISTINCT cmdb.classname as class1, cmdb2.classname as class2 FROM cmdbclasscatalog cmdb LEFT JOIN pg_catalog.pg_inherits cat ON cmdb.classid = cat.inhrelid LEFT JOIN cmdbclasscatalog cmdb2 ON cmdb2.classid = cat.inhparent WHERE NOT cmdb2.classname = \'\'')
        rows= cursor.fetchall()
        for r in rows:
            if r[0] not in self.classes:
                self.classes[r[0]] = {}
            if 'class_parents' not in self.classes[r[0]]:
                self.classes[r[0]]['class_parents'] = []
            self.classes[r[0]]['class_parents'].append(r[1])


    def __call__(self, cls, attr):
        return self.classes[cls][attr]

query_class_catalog = QueryClassCatalog()




"""
CMDBuild relies on the Shark workflow engine.  Each instance of the
Activity class, or one of its subclasses, represents a workflow process.
Therefore, CMDBuild needs to synchronize its own data model with the
engine's.  Interaction between the two takes place in these ways:

From CMDBuild to Shark
----------------------

1. The user alters a card for an Activity instance
#. CMDBuild stores the database change and invokes the Shark engine
#. Shark knows the new data and acts accordingly

From Shark to CMDBuild
----------------------

1. The XPDL workflow schema calls for CMDBuild-side actions, e.g.
changing an attribute's value, or advancing a process
#. Shark reads the XPDL file and notices that there is a "tool"
invoking a CMDBuild action
#. Shark calls a Java method that handles interaction with CMDBuild
#. The method prepares the data with JSON and calls the remote API
servlet
#. The servlet reads the data
#. The servlet changes the card and, in turns, invokes Shark as needed
(see above)
"""

import httplib
from cStringIO import StringIO

from django.utils import simplejson as json
from django_cmdbuild.serializer.jsonconverter import decorate, strip, ServerException

class Remote(object):
	class EmptyResponse(Exception):
		pass
	class GenericApplicationFailure(Exception):
		pass
	"""
	This class handles a remote connection towards CMDBuild's engine.
	"""
	def __init__(self, url=None):
		"""
		Prepare the connection object and connect to the server
		"""
		if url is None:
			from django.conf import settings
			url = settings.CMDBUILD_REMOTEAPI_URL
		self.proto, siteurl = url.split('://', 1)
		self.host, self.url = siteurl.split('/', 1)
		self.url = '/' + self.url
		self.conn = httplib.HTTPConnection(self.host)
		self.conn.connect()

	def request(self, command, data):
		"""
		Encode the data in CMDBuild's JSON dialect and send the command,
		then return the decoded response.
		"""
		import logging
		data = json.dumps(decorate(data))
		self.conn.request('POST', self.url, data, {
			'Content-Type': 'text/plain',
			'command': command
			})
		answer = self.conn.getresponse().read()
		if answer.isspace():
			raise self.EmptyResponse
		data = strip(json.loads(answer))
		if isinstance(data, ServerException):
			raise data
		return data


from django_cmdbuild.models.helpers import ClassFields, CMDBModelOptions, LookupField
from django_cmdbuild.models.querysets import *
from django_cmdbuild.remoteapi import http

class CMDBActivityOptions(CMDBModelOptions):

    class ReadOnly(Exception):
        """
        NEVER save an Activity instance directly -- it would get out
        of sync with respect to the Shark workflow.
        """
        pass

    def save(self):
        "Ban any attempt to save the instance."
        raise self.ReadOnly, 'Cannot save an Activity class'

    def _get_history(self):
        """
        Return a QuerySet with all cards belonging to this flow instance.
        """
        return self.__class__.objects.with_inactive().filter(
            process_code=self.process_code).order_by('begin_date')
    history = property(_get_history)

    def _get_flow_begin_date(self):
        """
        Return this flow's begin date.
        """
        return self.history[0].begin_date
    flow_begin_date = property(_get_flow_begin_date)

    def _get_flow_status_code(self):
        return Lookup.objects.get(type='FlowStatus',id=self.flow_status)

    def _is_stopped(self):
        return self.flow_status == Lookup.objects.get(type='FlowStatus', code='Interrotto')
    stopped = property(_is_stopped)
    
    def _is_started(self):
        return self.flow_status == Lookup.objects.get(type='FlowStatus', code='Avviato')
    started = property(_is_started)
        
    def _is_completed(self):
        return self.flow_status == Lookup.objects.get(type='FlowStatus', code='Completato')
    completed = property(_is_completed)

    def update_attribute(self, **attrmap):
        """
        Update an attribute of this class (with ``attname=value`` syntax).
        """
        remote = http.Remote()
        args = [self._meta.db_table, self.pk] + [self._make_bean(k, v)
            for k, v in attrmap.items()]
        resp = remote.request('card.update.id', args)
        if resp is not True:
            raise remote.GenericApplicationException(req)
        self._reload()

    @classmethod
    def create(cls, **kwargs):
        """
        Ask CMDBuild to start a new instance of this Activity.  If any
        attributes are provided, set them.
        """
        remote = http.Remote()
        table = cls._meta.db_table
        resp = remote.request('workflow.process.start', [table,
            cls._get_cmdbuild_attributemap(kwargs)])
        return cls.objects.get(process_code=resp)

    def advance(self, **kwargs):
        """
        Ask CMDBuild to let this activity step ahead, optionally updating
        its attributes.
        """
        remote = http.Remote()
        table = self._meta.db_table
        reply = remote.request('workflow.process.update',
            [self._meta.db_table, self.pk,
            self._get_cmdbuild_attributemap(kwargs)])
        self._reload()
        return reply


class ActivityFieldsQuerySet(ClassFieldsQuerySet):
    def _filter_by_flow_status(self, tag):
        from django_cmdbuild.models import Lookup
        l = Lookup.objects.get(type='FlowStatus', description=tag)
        return self.filter(flow_status=l.id)
    def completed(self):
        return self._filter_by_flow_status('Completato')
    def started(self):
        return self._filter_by_flow_status('Avviato')
    def stopped(self):
        return self._filter_by_flow_status('Interrotto')

class ActivityManager(QSManager(ActivityFieldsQuerySet)):
    def create(self, **kwargs):
        """
        Ask CMDBuild to start a new instance of this Activity.  If any
        attributes are provided, set them.
        """
        remote = http.Remote()
        table = self.model._meta.db_table
        resp = remote.request('workflow.process.start', [table,
            self.model._get_cmdbuild_attributemap(kwargs)])
        return self.model.objects.get(process_code=resp)

class ActivityFields(ClassFields):
    from django_cmdbuild.models import Lookup
    flow_status = LookupField('FlowStatus', db_column='FlowStatus')
    priority = models.IntegerField(db_column='Priority', null=True, blank=True)
    activity_definition = models.CharField(max_length=200, db_column='ActivityDefinitionId')
    process_code = models.CharField(max_length=200, db_column='ProcessCode')
    quick_accept = models.BooleanField(db_column='IsQuickAccept')
    activity_description = models.TextField(db_column='ActivityDescription', blank=True)
    objects = ActivityManager()




class MaskingDescriptor(object):
    def __init__(self, orig_desc):
        self.orig_desc = orig_desc
    # __get__ must modify the returned manager
    def __get__(self, *args, **kwargs):
        orig_manager = self.orig_desc.__get__(*args, **kwargs)
        class CMDBMapManager(type(orig_manager)):
            def __init__(self, orig):
                for k, v in orig.__dict__.items():
                    self.__dict__[k] = v
            def _add_items(self, source_col_name, target_col_name, *objs):
                # join_table: name of the m2m link table
                # source_col_name: the PK colname in join_table for the source object
                # target_col_name: the PK colname in join_table for the target object
                # *objs - objects to add. Either object instances, or primary keys of object instances.

                # If there aren't any objects, there is nothing to do.
                if objs:
                    # Check that all the objects are of the right type
                    new_ids = set()
                    for obj in objs:
                        if isinstance(obj, self.model):
                            new_ids.add(obj._get_pk_val())
                        else:
                            new_ids.add(obj)
                    # Add the newly created or already existing objects to the join table.
                    # First find out which items are already added, to avoid adding them twice
                    from django.db import connection
                    cursor = connection.cursor()
                    cursor.execute("SELECT %s FROM %s WHERE %s = %%s AND %s IN (%s)" % \
                        (target_col_name, self.join_table, source_col_name,
                        target_col_name, ",".join(['%s'] * len(new_ids))),
                        [self._pk_val] + list(new_ids))
                    existing_ids = set([row[0] for row in cursor.fetchall()])

                    # Add the ones that aren't there already
                    sql = ("INSERT INTO %s (\"IdDomain\", \"IdClass1\", \"IdClass2\", \"Status\", %s, %s) "
                        "VALUES ('%s', '\"%s\"', '\"%s\"', 'A', %%s, %%s)" % \
                        (self.join_table, source_col_name, target_col_name, self.join_table,
                        self.instance._meta.db_table, self.model._meta.db_table))
                    for obj_id in (new_ids - existing_ids):
                        cursor.execute(sql,
                            [self._pk_val, obj_id])
                    from django.db import transaction
                    transaction.commit_unless_managed()
        return CMDBMapManager(orig_manager)
            
from django.db import models

class CMDBManyToManyField(models.ManyToManyField):
    def __init__(self, *args, **kwargs):
        kwargs['limit_choices_to'] = {'status__exact': 'A'}
        super(CMDBManyToManyField, self).__init__(*args, **kwargs)
    def _get_m2m_column_name(self, related):
        return 'IdObj1'
    def _get_m2m_reverse_name(self, related):
        return 'IdObj2'
    def contribute_to_class(self, cls, name):
        super(CMDBManyToManyField, self).contribute_to_class(cls, name)
        desc = cls.__dict__[name]
        setattr(cls, name, MaskingDescriptor(desc))
    def contribute_to_related_class(self, cls, related):
        super(CMDBManyToManyField, self).contribute_to_related_class(cls, related)
        try:
            name = related.get_accessor_name()
            if name:
                try:
                    desc = cls.__dict__[name]
                    setattr(cls, name, MaskingDescriptor(desc))
                except KeyError:
                    pass
        except AttributeError:
            pass




from django.db import models

from django_cmdbuild.models.querysets import QSManager, ClassFieldsQuerySet

class LookUpManager(QSManager(ClassFieldsQuerySet)):
    def type(self, typename):
        "Returns all Lookup objects of the given type."
        return self.active().filter(type__exact=typename)

    def choices(self, typename):
        """
        Returns all Lookup objects of the given type, in a format
        suitable for usage as the ``choices`` keyword argument to
        a Django field.
        """
        rows = self.type(typename).values('id', 'description')
        return [(r['id'], r['description']) for r in rows]

    def get_label(self, typename, label):
        "Returns a single Lookup object w/ the given type and label."
        return self.get(type=typename, description=label)

    def get_by_number(self, typename, number):
        "Returns a single Lookup object of the given type and number."
        return self.get(type=typename, number=number)


class Lookup(models.Model):
      objects = LookUpManager()
      id = models.IntegerField(primary_key=True, db_column='Id', blank=True, help_text=u'Topic', null=True)
      type = models.CharField(help_text=u'Topic', max_length=32, blank=True, db_column='Type')
      parenttype = models.CharField(help_text=u'Topic', max_length=32, blank=True, db_column='ParentType')
      parentid = models.IntegerField(help_text=u'Topic', blank=True, null=True, db_column='ParentId')
      number = models.IntegerField(help_text=u'Topic', blank=True, null=True, db_column='Number')
      code = models.CharField(help_text=u'Topic', max_length=100, blank=True, db_column='Code')
      description = models.CharField(help_text=u'Topic', max_length=100, blank=True, db_column='Description')
      isdefault = models.BooleanField(help_text=u'Topic', blank=True, null=True, db_column='IsDefault')
      status = models.TextField(help_text=u'Topic', blank=True, db_column='Status') # This field type is a guess.
      notes = models.TextField(help_text=u'Topic', blank=True, db_column='Notes')
      class Admin:
          list_display = ('type', 'description')
      class Meta:
          db_table = u'LookUp'
      def __unicode__(self):
          return self.description



from setuptools import setup, find_packages
setup(
    name = "dj-cmdbuild",
    version = "0.1",
    packages = find_packages(exclude=['._*']),
    exclude_package_data = { '': ['*/local_*.py', '*/._*']},

    author = "Emanuele Pucciarelli",
    author_email = "ep@acm.org",
    description = "Tools for interfacing Django with CMDBuild databases",
    keywords = "django cmdbuild",
    url = "http://code.google.com/p/dj-cmdbuild",
    license = "GPL",

)



from django.core.management.base import NoArgsCommand, CommandError

def log(string):
    print '# %s' % string

class Field(object):
    def __init__(self, name, klass=None, param=None):
        self.name = name
        self.kw_params = {}
        if param:
            self.params = [param]
        else:
            self.params = []
        self.comments = []
        if klass:
            self.set_class(klass)
        self.mangle_name()
    def add_kw_param(self, name, value, representation=True):
        if representation:
            value = '%r' % value
        self.kw_params[name] = value
    def add_param(self, value, representation=True):
        if representation:
            value = '%r' % value
        self.params.append(value)
    def add_comment(self, comment):
        self.comments.append(comment)
    def set_class(self, klass):
        self.klass = klass
    def line(self):
        if hasattr(self, 'db_column'):
            self.add_kw_param('db_column', self.db_column)
        params = self.params + ['%s=%s' % (k, v) for k, v in self.kw_params.items()] 
        l = '    %s = %s(%s)' % (self.name, self.klass, ', '.join(params))
        if self.comments:
            l += ' # ' + ' '.join(self.comments)
        yield l
    def mangle_name(self):
        if ' ' in self.name:
            self.name = att_name.replace(' ', '')
            self.add_comment('Field renamed to remove spaces.')
        import keyword
        if keyword.iskeyword(self.name):
            self.name += '_field'
            self.add_comment('Field renamed because it was a Python reserved word.')
        
        pass
    def set_column(self, column_name):
        self.db_column = column_name
            
class Vertex(object):
    def __init__(self, name, graph=None):
        self.name = name
        self.store = graph
        self.edge_out = {}
        self.edge_in = {}
    def add_edge(self, name):
        if name not in self.edge_out:
            self.edge_out[name] = 1
            self.store.vertex(name).edge_in[self.name] = 1
    def remove_edge(self, name):
        del self.edge_out[name]
        del self.store.vertex(name).edge_in[self.name]
    def has_outgoing_edges(self):
        return len(self.edge_out.keys()) > 0
    def has_incoming_edges(self):
        return len(self.edge_in.keys()) > 0

table2model = lambda table_name: table_name.title().replace('_', '')

class ModelClass(Vertex):
    def __init__(self, name, store=None, graph=None):
        Vertex.__init__(self, name, graph=graph)
        self.base_classes = ['models.Model']
        self.lines = []
        self.extra_attributes = []
        self.fields = []
        self.name = name
        self.store = store
    def add_base_class(self, name):
        self.base_classes.append(name)
    def add(self, line):
        self.lines.append(line)
    def extra(self, line):
        self.add('      %s' % line)
    def add_field(self, field):
        self.fields.append(field)
    def yield_self(self):
        if not len(self.lines):
            return
        yield 'class %s(%s):' % (table2model(self.name), ', '.join(self.base_classes))
        for f in self.fields:
            for l in f.line():
                yield l
        yield '      class Admin:'
        if 'ClassFields' in self.base_classes:
            yield "          list_display = ('description', 'code', 'status')"
        else:
            yield "          pass"
        for l in self.lines:
            yield l
        raise StopIteration
        
class ModelStore(object):
    def __init__(self):
        self.models = {}
        self.bases = {}
    def vertex(self, name):
        return self.model(name)
    def model(self, name):
        try:
            return self.models[name]
        except KeyError:
            self.models[name] = ModelClass(name, store=self)
            return self.models[name]
    def delete(self, name):
        del(self.models[name])
    def with_classfields(self):
        self.bases['Class'] = 1
        self.bases['ExpanderField'] = 1
    def with_activityfields(self):
        self.bases['Activity'] = 1
        self.bases['ExpanderField'] = 1
    def with_lookup(self):
        self.bases['LookUp'] = 1
    def with_mapfields(self):
        self.bases['MapFields'] = 1
    def with_manytomany(self):
        self.bases['CMDBManyToManyField'] = 1
    def sorted_models(self):
        q = filter(lambda x: not x.has_incoming_edges(), self.models.values())
        l = []
        while len(q) > 0:
            model = q.pop()
            log('Processing %s' % model.name)
            l.append(model)
            for needed in model.edge_out.keys():
                log( ' %s depends on %s, removing' % (model.name, needed))
                model.remove_edge(needed)
                if not self.model(needed).has_incoming_edges():
                    log('  %s has nothing else depending on it, queueing' % needed)
                    q.append(self.model(needed))
                else:
                    log('  %s still has %s depending on it' % (needed, self.model(needed).edge_in.keys()))
        if len(filter(lambda x: x.has_incoming_edges(), self.models.values())) > 0:
            for m in self.models.values():
                if m.has_incoming_edges():
                    m.extra('# Caught in a cycle!')
                    l.append(m)
        for b in self.bases.keys():
            if b == 'Class':
                yield 'from django_cmdbuild.models.helpers import ClassFieldsManager, ClassFields'
            elif b == 'Activity':
                yield 'from django_cmdbuild.models.helpers import ActivityFields'
            elif b == 'LookUp':
                yield "from django_cmdbuild.models.helpers import LookUpManager"
            elif b == 'CMDBManyToManyField':
                yield "from django_cmdbuild.models.related import %s" % b
            elif b == 'ExpanderField':
                yield "from django_cmdbuild.models.expander import %s" % b
            else:
                yield "from django_cmdbuild.models.helpers import %s" % b
        for m in l:
            yield m

from django_cmdbuild.models.introspection_common import query_class_catalog

class Command(NoArgsCommand):
    help = "Introspects the database tables in the given database and outputs a Django model module."

    requires_model_validation = False

    def handle_noargs(self, **options):
        try:
            for line in self.handle_inspection():
                print line
        except NotImplementedError:
            raise CommandError("Database inspection isn't supported for the currently selected database backend.")
    
    
    def output_classfields(self):
        yield 'AAA'
        if not getattr(self, 'classfields_are_out'):
            print 'BBB'
            yield ''
            self.classfields_are_out = True

    def __init__(self, *args, **kwargs):
        super(Command, self).__init__(*args, **kwargs)
        self._dom_cache = None

    def query_dom_catalog(self, name):
        if self._dom_cache is None:
            self._dom_cache = {}
            from django.db import connection
            cursor = connection.cursor()
            cursor.execute('SELECT domainname, domainclass1, domainclass2, domaincardinality '
                'FROM cmdbdomaincatalog')
            for row in cursor.fetchall():
                self._dom_cache[row[0]] = {'name': row[0], 'class1': row[1],
                    'class2': row[2], 'cardinality': row[3]}
        try:
            return self._dom_cache[name]
        except IndexError:
            raise KeyError('No catalog entry for table %s' % name)
    
    def handle_inspection(self):
        from django.db import connection, get_introspection_module
        import keyword

        introspection_module = get_introspection_module()

        def maptable2attr(name):
            name = name.lower()
            if name.startswith('map_'):
                name = name[4:]
            return name + 's'

        cursor = connection.cursor()
        yield "# This is an auto-generated Django model module."
        yield "# You'll have to do the following manually to clean this up:"
        yield "#     * Rearrange models' order"
        yield "#     * Make sure each model has one field with primary_key=True"
        yield "# Feel free to rename the models, but don't rename db_table values or field names."
        yield "#"
        yield "# Also note: You'll have to insert the output of 'django-admin.py sqlcustom [appname]'"
        yield "# into your database."
        yield ''
        yield 'from django.db.models import *'
        yield 'from django.db import models'
        yield ''
        
        store = ModelStore()
        
        for table_name in introspection_module.get_table_list(cursor):
            # Internal "service" views
            if table_name.startswith('cmdb') and table_name.endswith('catalog'):
                continue
            if table_name in ('relationhistorylist', 'relationlist'):
                continue 
            # History views are handled separately
            if table_name.endswith('_history'):
                continue
            # LookUp has its own built-in model
            if table_name == 'LookUp':
                continue
            mc = store.model(table_name)
            # Special CMDBuild cases
            skip_rows = []
            extra_attributes = []
            names = []
            # Collect
            for row in introspection_module.get_table_description(cursor, table_name):
                names.append(row[0].lower())
            def subset(small, big):
                bools= map(lambda x: x in big, small)
                excluder = filter(lambda y: y != True, bools)
                return len(excluder) == 0

            activity_relations = ('flowstatus', 'priority',
                'activitydefinitionid', 'processcode', 'isquickaccept',
                'activitydescription')
            class_relations = ('id', 'code', 'description', 'status',
                'user', 'begindate', 'idclass')
            
            if subset(class_relations, names):
                skip_rows.extend(class_relations)
                if subset(activity_relations, names):
                    skip_rows.extend(activity_relations)
                    mc.add_base_class('ActivityFields')
                    mc.add_base_class('CMDBActivityOptions')
                    mc.extra('activityfields = ExpanderField(ActivityFields)')
                    store.with_activityfields()
                else:
                    mc.add_base_class('ClassFields')
                    mc.add_base_class('CMDBModelOptions')
                    mc.extra('commonfields = ExpanderField(ClassFields)')
                    store.with_classfields()
                

            map_relations = ('iddomain', 'idclass1', 'idclass2', 'status',
                'user', 'begindate', 'enddate')

            if subset(map_relations, names):
                try:
                    cat_info = self.query_dom_catalog(mc.name)
                    if cat_info['cardinality'] == 'N:N':
                        store.with_manytomany()
                        first_table = store.model(cat_info['class1'])
                        first_table.extra('%s = CMDBManyToManyField(%s, db_table=\'%s\', blank=True, null=True)' % (
                            maptable2attr(table_name),
                            cat_info['class2'],
                            table_name))
                        store.model(cat_info['class2']).add_edge(first_table.name)
                        store.delete(mc.name)
                        continue
                    else:
                        mc.extra('# Links %s and %s with cardinality %s' %( cat_info['class1'],
                            cat_info['class2'], cat_info['cardinality']))
                        mc.extra('# Foreign keys should be available in the objects')
                except KeyError:
                    log('%s not found' % mc.name)
                    pass
                skip_rows.extend(map_relations)
                mc.add_base_class('MapFields')
                mc.extra('mapfields = ExpanderField(MapFields)')
                store.with_mapfields()

            try:
                relations = introspection_module.get_relations(cursor, table_name)
            except NotImplementedError:
                relations = {}
            try:
                indexes = introspection_module.get_indexes(cursor, table_name)
            except NotImplementedError:
                indexes = {}
            
            log('%s: %s' % (table_name, indexes))
            try:
                log ('inheritance: %s %s' % (table_name, query_class_catalog(table_name, 'class_parents')))
            except KeyError:
                pass
            
            for i, row in enumerate(introspection_module.get_table_description(cursor, table_name)):
                log('%s' % (row,))
                att_name = row[0].lower()
                if att_name in skip_rows:
                    continue
                f = Field(att_name)
                comment_notes = [] # Holds Field notes, to be displayed in a Python comment.
                extra_params = {}  # Holds Field parameters such as 'db_column'.
                try:
                    cat_entry = query_class_catalog(mc.name, row[0])
                except KeyError:
                    #cat_entry = None
                    log('%s, %s not found' % (mc.name, row[0]))
                    
                f.set_column(row[0])
                # extra_params['db_column'] = '%r' % column_name
                # Special treatment for idclass
                #if f.name == 'idclass':
                #    f.name = '_' + f.name
                #    f.add_kw_param('editable', False)
                #    f.add_kw_param('default', '"%s"' % table_name)
                if f.name == 'code':
                    f.set_class('CodeField')
                elif f.name == 'description':
                    f.set_class('DescriptionField')
                elif i in relations:
                    #mc.add_edge(relations[i][1])
                    store.model(relations[i][1]).add_edge(mc.name)
                    rel_to = relations[i][1] == table_name and "'self'" or table2model(relations[i][1])
                    #field_type = 'ForeignKey(%s' % rel_to
                    f.set_class('models.ForeignKey')
                    f.add_param(rel_to, representation=False)
                    if f.name.endswith('_id'):
                        f.name = f.name[:-3]
                else:
                    try:
                        field_type = introspection_module.DATA_TYPES_REVERSE[row[1]] 
                    except KeyError:
                        field_type = 'TextField'
                        f.add_comment('This field type is a guess.')
                    f.set_class('models.' + field_type)


                    # This is a hook for DATA_TYPES_REVERSE to return a tuple of
                    # (field_type, extra_params_dict).
                    if type(field_type) is tuple:
                        field_type, new_params = field_type
                        for k, v in new_params.items():
                            f.add_kw_param(k, v)
                        extra_params.update(new_params)

                    # Add max_length for all CharFields.
                    if field_type == 'CharField' and row[3]:
                        f.add_kw_param('max_length', row[3])

                    if field_type == 'DecimalField':
                        f.add_kw_param('max_digits', row[4])
                        f.add_kw_param('decimal_places', row[5])
                        #extra_params['max_digits'] = '%r' % row[4]
                        #extra_params['decimal_places'] = '%r' % row[5]
                    
                    if cat_entry:
                        # Add help text.
                        if cat_entry['desc']:
                            f.add_kw_param('help_text', cat_entry['desc'])
                            #extra_params['help_text'] = '%r' % cat_entry['desc']
                        if ( len(cat_entry['lookup']) > 0 
                            and (not table_name == 'Lookup')):
                            #extra_params['choices'] = 'Lookup.objects.choices(%r)' % cat_entry['lookup']
                            #f.add_kw_param('choices', 'Lookup.objects.choices(%r)' % cat_entry['lookup'],
                                #representation=False)
                            #store.model('LookUp').add_edge(mc.name)
                            f.set_class('LookupField')
                            f.add_param(cat_entry['lookup'])
                            #store.with_lookup()

                    # Add primary_key and unique, if necessary.
                    if f.db_column in indexes:
                        if indexes[f.db_column]['primary_key']:
                            #extra_params['primary_key'] = 'True'
                            f.add_kw_param('primary_key', True)
                        elif indexes[f.db_column]['unique']:
                            #extra_params['unique'] = 'True'
                            f.add_kw_param('unique', True)

                    field_type += '('

                # Don't output 'id = meta.AutoField(primary_key=True)', because
                # that's assumed if it doesn't exist.
                if att_name == 'id' and field_type == 'AutoField(' and extra_params == {'primary_key': True}:
                    continue

                # Add 'null' and 'blank', if the 'null_ok' flag was present in the
                # table description.
                try:
                    if cat_entry['null']: # If it's NULL...
                        f.add_kw_param('blank', True)
                        if not field_type in ('TextField(', 'CharField('):
                            f.add_kw_param('null', True)
                except TypeError:
                    pass

                #field_desc = '%s = models.%s' % (att_name, field_type)
                #if extra_params:
                #    if not field_desc.endswith('('):
                #        field_desc += ', '
                #    field_desc += ', '.join(['%s=%s' % (k, v) for k, v in extra_params.items()])
                #field_desc += ')'
                #if comment_notes:
                    #field_desc += ' # ' + ' '.join(comment_notes)
                #mc.add( '      %s' % field_desc)
                mc.add_field(f)
            mc.add( '      class Meta:')
            mc.add( '          db_table = %r' % table_name)
            mc.add( '')
            #for l in mc.yield_self():
            #   yield l
        log(store.models)
        for m in store.sorted_models():
            if isinstance(m, str):
                yield m
            else:
                for l in m.yield_self():
                    yield l

def inspect():
    Command().handle_noargs()
            
if __name__ == '__main__':
    inspect()
